@@ -1,5 +1,5 @@
 from binascii import hexlify
-from datetime import datetime
+from datetime import datetime, timezone
 from getpass import getuser
 from itertools import groupby
 import errno
@@ -16,9 +16,10 @@ import sys
 import time
 from io import BytesIO
 from . import xattr
-from .helpers import parse_timestamp, Error, uid2user, user2uid, gid2group, group2gid, format_timedelta, \
+from .helpers import Error, uid2user, user2uid, gid2group, group2gid, \
+    parse_timestamp, to_localtime, format_time, format_timedelta, \
     Manifest, Statistics, decode_dict, make_path_safe, StableDict, int_to_bigint, bigint_to_int, \
-    st_atime_ns, st_ctime_ns, st_mtime_ns, ProgressIndicatorPercent
+    ProgressIndicatorPercent
 from .platform import acl_get, acl_set
 from .chunker import Chunker
 from .hashindex import ChunkIndex
@@ -26,10 +27,10 @@ import msgpack
 
 ITEMS_BUFFER = 1024 * 1024
 
-CHUNK_MIN_EXP = 10  # 2**10 == 1kiB
+CHUNK_MIN_EXP = 19  # 2**19 == 512kiB
 CHUNK_MAX_EXP = 23  # 2**23 == 8MiB
 HASH_WINDOW_SIZE = 0xfff  # 4095B
-HASH_MASK_BITS = 16  # results in ~64kiB chunks statistically
+HASH_MASK_BITS = 21  # results in ~2MiB chunks statistically
 
 # defaults, use --chunker-params to override
 CHUNKER_PARAMS = (CHUNK_MIN_EXP, CHUNK_MAX_EXP, HASH_MASK_BITS, HASH_WINDOW_SIZE)
@@ -37,17 +38,11 @@ CHUNKER_PARAMS = (CHUNK_MIN_EXP, CHUNK_MAX_EXP, HASH_MASK_BITS, HASH_WINDOW_SIZE
 # chunker params for the items metadata stream, finer granularity
 ITEMS_CHUNKER_PARAMS = (12, 16, 14, HASH_WINDOW_SIZE)
 
-utime_supports_fd = os.utime in getattr(os, 'supports_fd', {})
-utime_supports_follow_symlinks = os.utime in getattr(os, 'supports_follow_symlinks', {})
-has_mtime_ns = sys.version >= '3.3'
 has_lchmod = hasattr(os, 'lchmod')
 has_lchflags = hasattr(os, 'lchflags')
 
-# Python <= 3.2 raises OSError instead of PermissionError (See #164)
-try:
-    PermissionError = PermissionError
-except NameError:
-    PermissionError = OSError
+flags_normal = os.O_RDONLY | getattr(os, 'O_BINARY', 0)
+flags_noatime = flags_normal | getattr(os, 'O_NOATIME', 0)
 
 
 class DownloadPipeline:
@@ -137,7 +132,7 @@ class Archive:
     def __init__(self, repository, key, manifest, name, cache=None, create=False,
                  checkpoint_interval=300, numeric_owner=False, progress=False,
                  chunker_params=CHUNKER_PARAMS,
-                 start=datetime.now(), end=datetime.now()):
+                 start=datetime.utcnow(), end=datetime.utcnow()):
         self.cwd = os.getcwd()
         self.key = key
         self.repository = repository
@@ -181,14 +176,22 @@ class Archive:
     def load(self, id):
         self.id = id
         self.metadata = self._load_meta(self.id)
-        decode_dict(self.metadata, (b'name', b'hostname', b'username', b'time'))
+        decode_dict(self.metadata, (b'name', b'hostname', b'username', b'time', b'time_end'))
         self.metadata[b'cmdline'] = [arg.decode('utf-8', 'surrogateescape') for arg in self.metadata[b'cmdline']]
         self.name = self.metadata[b'name']
 
     @property
     def ts(self):
-        """Timestamp of archive creation in UTC"""
-        return parse_timestamp(self.metadata[b'time'])
+        """Timestamp of archive creation (start) in UTC"""
+        ts = self.metadata[b'time']
+        return parse_timestamp(ts)
+
+    @property
+    def ts_end(self):
+        """Timestamp of archive creation (end) in UTC"""
+        # fall back to time if there is no time_end present in metadata
+        ts = self.metadata.get(b'time_end') or self.metadata[b'time']
+        return parse_timestamp(ts)
 
     @property
     def fpr(self):
@@ -196,15 +199,19 @@ class Archive:
 
     @property
     def duration(self):
-        return format_timedelta(self.end-self.start)
+        return format_timedelta(self.end - self.start)
 
     def __str__(self):
-        return '''Archive name: {0.name}
+        return '''\
+Archive name: {0.name}
 Archive fingerprint: {0.fpr}
-Start time: {0.start:%c}
-End time: {0.end:%c}
+Time (start): {start}
+Time (end):   {end}
 Duration: {0.duration}
-Number of files: {0.stats.nfiles}'''.format(self)
+Number of files: {0.stats.nfiles}'''.format(
+            self,
+            start=format_time(to_localtime(self.start.replace(tzinfo=timezone.utc))),
+            end=format_time(to_localtime(self.end.replace(tzinfo=timezone.utc))))
 
     def __repr__(self):
         return 'Archive(%r)' % self.name
@@ -235,7 +242,13 @@ Number of files: {0.stats.nfiles}'''.format(self)
             raise self.AlreadyExists(name)
         self.items_buffer.flush(flush=True)
         if timestamp is None:
-            timestamp = datetime.utcnow()
+            self.end = datetime.utcnow()
+            start = self.start
+            end = self.end
+        else:
+            self.end = timestamp
+            start = timestamp
+            end = timestamp  # we only have 1 value
         metadata = StableDict({
             'version': 1,
             'name': name,
@@ -243,7 +256,8 @@ Number of files: {0.stats.nfiles}'''.format(self)
             'cmdline': sys.argv,
             'hostname': socket.gethostname(),
             'username': getuser(),
-            'time': timestamp.isoformat(),
+            'time': start.isoformat(),
+            'time_end': end.isoformat(),
         })
         data = msgpack.packb(metadata, unicode_errors='surrogateescape')
         self.id = self.key.id_hash(data)
@@ -301,7 +315,7 @@ Number of files: {0.stats.nfiles}'''.format(self)
             else:
                 os.unlink(path)
         except UnicodeEncodeError:
-            raise self.IncompatibleFilesystemEncodingError(path, sys.getfilesystemencoding())
+            raise self.IncompatibleFilesystemEncodingError(path, sys.getfilesystemencoding()) from None
         except OSError:
             pass
         mode = item[b'mode']
@@ -341,7 +355,7 @@ Number of files: {0.stats.nfiles}'''.format(self)
             try:
                 os.symlink(source, path)
             except UnicodeEncodeError:
-                raise self.IncompatibleFilesystemEncodingError(source, sys.getfilesystemencoding())
+                raise self.IncompatibleFilesystemEncodingError(source, sys.getfilesystemencoding()) from None
             self.restore_attrs(path, item, symlink=True)
         elif stat.S_ISFIFO(mode):
             if not os.path.exists(os.path.dirname(path)):
@@ -392,12 +406,10 @@ Number of files: {0.stats.nfiles}'''.format(self)
         else:
             # old archives only had mtime in item metadata
             atime = mtime
-        if fd and utime_supports_fd:  # Python >= 3.3
+        if fd:
             os.utime(fd, None, ns=(atime, mtime))
-        elif utime_supports_follow_symlinks:  # Python >= 3.3
+        else:
             os.utime(path, None, ns=(atime, mtime), follow_symlinks=False)
-        elif not symlink:
-            os.utime(path, (atime / 1e9, mtime / 1e9))
         acl_set(path, item, self.numeric_owner)
         # Only available on OS X and FreeBSD
         if has_lchflags and b'bsdflags' in item:
@@ -441,9 +453,9 @@ Number of files: {0.stats.nfiles}'''.format(self)
             b'mode': st.st_mode,
             b'uid': st.st_uid, b'user': uid2user(st.st_uid),
             b'gid': st.st_gid, b'group': gid2group(st.st_gid),
-            b'atime': int_to_bigint(st_atime_ns(st)),
-            b'ctime': int_to_bigint(st_ctime_ns(st)),
-            b'mtime': int_to_bigint(st_mtime_ns(st)),
+            b'atime': int_to_bigint(st.st_atime_ns),
+            b'ctime': int_to_bigint(st.st_ctime_ns),
+            b'mtime': int_to_bigint(st.st_mtime_ns),
         }
         if self.numeric_owner:
             item[b'user'] = item[b'group'] = None
@@ -535,7 +547,7 @@ Number of files: {0.stats.nfiles}'''.format(self)
         item = {b'path': safe_path}
         # Only chunkify the file if needed
         if chunks is None:
-            fh = Archive._open_rb(path, st)
+            fh = Archive._open_rb(path)
             with os.fdopen(fh, 'rb') as fd:
                 chunks = []
                 for chunk in self.chunker.chunkify(fd, fh):
@@ -557,53 +569,22 @@ Number of files: {0.stats.nfiles}'''.format(self)
             yield Archive(repository, key, manifest, name, cache=cache)
 
     @staticmethod
-    def _open_rb(path, st):
-        flags_normal = os.O_RDONLY | getattr(os, 'O_BINARY', 0)
-        flags_noatime = flags_normal | getattr(os, 'O_NOATIME', 0)
-        euid = None
-
-        def open_simple(p, s):
-            return os.open(p, flags_normal)
-
-        def open_noatime(p, s):
-            return os.open(p, flags_noatime)
-
-        def open_noatime_if_owner(p, s):
-            if euid == 0 or s.st_uid == euid:
-                # we are root or owner of file
-                return open_noatime(p, s)
-            else:
-                return open_simple(p, s)
-
-        def open_noatime_with_fallback(p, s):
-            try:
-                fd = os.open(p, flags_noatime)
-            except PermissionError:
-                # Was this EPERM due to the O_NOATIME flag?
-                fd = os.open(p, flags_normal)
-                # Yes, it was -- otherwise the above line would have thrown
-                # another exception.
-                nonlocal euid
-                euid = os.geteuid()
-                # So in future, let's check whether the file is owned by us
-                # before attempting to use O_NOATIME.
-                Archive._open_rb = open_noatime_if_owner
-            return fd
-
-        if flags_noatime != flags_normal:
-            # Always use O_NOATIME version.
-            Archive._open_rb = open_noatime_with_fallback
-        else:
-            # Always use non-O_NOATIME version.
-            Archive._open_rb = open_simple
-        return Archive._open_rb(path, st)
+    def _open_rb(path):
+        try:
+            # if we have O_NOATIME, this likely will succeed if we are root or owner of file:
+            return os.open(path, flags_noatime)
+        except PermissionError:
+            if flags_noatime == flags_normal:
+                # we do not have O_NOATIME, no need to try again:
+                raise
+            # Was this EPERM due to the O_NOATIME flag? Try again without it:
+            return os.open(path, flags_normal)
 
 
 # this set must be kept complete, otherwise the RobustUnpacker might malfunction:
 ITEM_KEYS = set([b'path', b'source', b'rdev', b'chunks',
                  b'mode', b'user', b'group', b'uid', b'gid', b'mtime', b'atime', b'ctime',
-                 b'xattrs', b'bsdflags', b'acl_nfs4', b'acl_access', b'acl_default', b'acl_extended',
-            ])
+                 b'xattrs', b'bsdflags', b'acl_nfs4', b'acl_access', b'acl_default', b'acl_extended', ])
 
 
 class RobustUnpacker:
@@ -863,7 +844,7 @@ class ArchiveChecker:
                 archive = StableDict(msgpack.unpackb(data))
                 if archive[b'version'] != 1:
                     raise Exception('Unknown archive metadata version')
-                decode_dict(archive, (b'name', b'hostname', b'username', b'time'))
+                decode_dict(archive, (b'name', b'hostname', b'username', b'time', b'time_end'))
                 archive[b'cmdline'] = [arg.decode('utf-8', 'surrogateescape') for arg in archive[b'cmdline']]
                 items_buffer = ChunkBuffer(self.key)
                 items_buffer.write_chunk = add_callback
